from __future__ import print_function
import sys
from atlas.cortex.server import WebServer
from atlas.cortex.server import McCortexQuery
from atlas.cortex.server import GraphWalker
from atlas.cortex.server import query_mccortex
from atlas.utils import get_params
import socket
import json
from pprint import pprint
import logging
from Bio import SeqIO
import argparse

sys.path.append('/home/phelimb/git/atlas-core')

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)

# Using seed kmers from a gene panel return a DFS through the graph

parser = argparse.ArgumentParser(description='Add length argument to panel')
parser.add_argument(
    'dna_fasta',
    metavar='dna_fasta',
    type=str,
    help='dna_fasta')
parser.add_argument(
    '--prot_fasta',
    metavar='prot_fasta',
    type=str,
    help='prot_fasta')
parser.add_argument(
    '-f',
    '--ctx',
    metavar='ctx',
    type=str,
    help='cortex graph binary')
parser.add_argument('-k', '--kmer_size', metavar='kmer_size', type=int,
                    help='kmer_size', default=31)
parser.add_argument('-p', '--port', metavar='port', type=int,
                    help='port', default=None)
args = parser.parse_args()


# def get_open_port():
# 	if args.port:
# 		return args.port
# 	else:
# 		s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
# 		s.bind(("",0))
# 		s.listen(1)
# 		port = s.getsockname()[1]
# 		s.close()
# 		return port

class PathDetails(object):

    def __init__(self, start_kmer, last_kmer, length, skipped=0, v=""):
        self.start_kmer = start_kmer
        self.last_kmer = last_kmer
        self.length = length
        self.skipped = skipped
        self.version = v
        self.repeat_kmers = {}

    def __eq__(self, pd):
        return self.start_kmer == pd.start_kmer and self.last_kmer == pd.last_kmer and self.length == pd.length and self.skipped == pd.skipped

    def set_repeat_kmers(self, repeat_kmers):
        if not self.repeat_kmers:
            self.repeat_kmers = repeat_kmers
        else:
            raise ValueError("Already set repeat kmers")

genes = {}
skip_list = {"tem": ["191", "192"],
             "oxa": ["12", "14", "33"],
             "shv": ["12", "6"]
             }


def get_repeat_kmers(record, k):
    # Process repeat kmers
    kmers = {}
    kmers_seq = []
    # repeat_kmers = {}
    for i in range(len(record.seq) - k + 1):
        _kmers = [str(record.seq[i:i + k]),
                  str(record.seq[i:i + k].reverse_complement())]
        for kmer in _kmers:
            kmers_seq.append(kmers_seq)
            if kmer in kmers:
                c = max(kmers[kmer].keys()) + 1
                kmers[kmer][c] = str(record.seq[i + 1:i + k + 1])
            else:
                kmers[kmer] = {}
                kmers[kmer][1] = str(record.seq[i + 1:i + k + 1])
    # repeat_kmers = {}
    # for kmer, count in kmers.items():
    # 	if len(count.keys()) > 1:
    # 		repeat_kmers[kmer] = count
    return kmers


def find_start_kmer(seq, mcq, k=args.kmer_size):
    skipped = 0
    for i in range(len(record.seq) - k + 1):
        kmer = seq[i:i + k]
        q = mcq.query(kmer)
        if q.data and q.depth > 1:
            return kmer, skipped
        skipped += 1
    return None, -1


def choose_best_assembly(paths):
    paths.sort(key=lambda x: x["min_non_zero_depth"], reverse=True)
    current_best = paths[0]
    for path in paths[1:]:
        if path["min_non_zero_depth"] < current_best["min_non_zero_depth"]:
            return current_best
        elif path["min_non_zero_depth"] == current_best["min_non_zero_depth"]:
            if path["median_depth"] > current_best["median_depth"]:
                current_best = path
    return current_best


def get_paths_for_gene(gene_name, gene_dict, gw):
    paths = {}
    current_prots = []
    for pd in gene_dict["pathdetails"]:
        path_for_pd = gw.breath_first_search(
            N=pd.length,
            seed=pd.start_kmer,
            end_kmers=[
                pd.last_kmer],
            known_kmers=gene_dict["known_kmers"],
            repeat_kmers=pd.repeat_kmers,
            N_left=pd.skipped)
        # print (p)
        if path_for_pd:
            keep_p = []
            for p in path_for_pd:
                if p["prot"] not in current_prots:
                    keep_p.append(p)
                    current_prots.append(p["prot"])
            if keep_p:
                if len(keep_p) > 1:
                    raise NotImplementedError()
                else:
                    paths[pd.version] = keep_p[0]
    return paths


wb = None
if args.port is None:
    if args.ctx is None:
        raise ValueError(
            "Require either port or binary. e.g. atlas contigs panel.fasta -f sample.ctx")
    else:
        # port =  (get_open_port())
        # logger.debug("Running server on port %i " % port)
        wb = WebServer(port=0, args=[args.ctx])
        logger.debug("Loading binary")
        wb.start()
        # Serve on a thread
        # logger.debug("Starting sever")
        # server = multiprocessing.Process(target=wb.serve, args = (1000,))
        # server.start()
else:
    port = args.port

logger.debug("Walking the graph")
out_dict = {}
gw = GraphWalker(proc=wb.mccortex, kmer_size=args.kmer_size, print_depths=True)


with open(args.dna_fasta, 'r') as infile:
    for i, record in enumerate(SeqIO.parse(infile, "fasta")):
        repeat_kmers = get_repeat_kmers(record, args.kmer_size)
        params = get_params(record.id)
        gene_name = params.get("name", i)
        version = params.get("version", i)
        # print (params, gene_name, version)

        # start_kmer = str(record.seq)[:args.kmer_size]
        last_kmer = str(record.seq)[-args.kmer_size:]
        start_kmer, skipped = find_start_kmer(
            str(record.seq), gw.mcq, args.kmer_size)
        if gene_name not in genes:
            genes[gene_name] = {}
            genes[gene_name]["pathdetails"] = []
            genes[gene_name]["known_kmers"] = ""
        if version not in skip_list.get(gene_name, []) and start_kmer:
            pd = PathDetails(start_kmer, last_kmer, len(record.seq),
                             skipped=skipped, v=version)
            pd.set_repeat_kmers(repeat_kmers)
            # print (gene_name, version, len(record.seq))
            # if pd not in genes[gene_name]["pathdetails"]:
            genes[gene_name]["pathdetails"].append(pd)
            # else:
            # j = genes[gene_name]["pathdetails"].index(pd)
            # _pd = genes[gene_name]["pathdetails"][j]
            # _pd.version += "-%s" % version
            # _pd.set_repeat_kmers(repeat_kmers)

        if gene_name in genes:
            genes[gene_name]["known_kmers"] += "%sN" % str(record.seq)


# logger.debug(len(genes.get("blaZ",{}).get("pathdetails")))
# with open("gene.tmp.json", "w") as outf:
# 	json.dump(genes, outf, sort_keys = False, indent = 4)

for gene_name, gene_dict in genes.items():
    paths = get_paths_for_gene(gene_name, gene_dict, gw)
    if len(paths.keys()) > 1:
        # choose best version
        best_path = choose_best_assembly(paths.values())
    elif len(paths.keys()) == 1:
        best_path = paths.values()[0]
    else:
        best_path = {}
    out_dict[gene_name] = best_path
print (json.dumps(out_dict, sort_keys=False, indent=4))
logger.info("Cleaning up")
if wb is not None:
    wb.stop()



import os
import hashlib
import re
import json


def check_args(args):
    if args.db_name is None:
        args.db_name = os.environ.get("DB_NAME")
    if args.db_name is None:
        raise ValueError(
            "db_name needs to be set. Either run with --db_name :db_name or export DB_NAME=:db_name")
    return args


def make_hash(s):
    return hashlib.sha256(s.encode("ascii", errors="ignore")).hexdigest()


def make_var_hash(ref, pos, alts):
    var = "".join(
        [ref, str(pos), "/".join(alts)])
    return make_hash(var)


def split_var_name(name):
    items = re.match(r"([A-Z]+)([-0-9]+)([A-Z/]+)", name, re.I).groups()
    return items[0], int(items[1]), items[2]


def unique(l):
    seen = set()
    return [x for x in l if x not in seen and not seen.add(x)]


def flatten(l):
    return [item for sublist in l for item in sublist]


def get_params(url):
    params = {}
    try:
        p_str = url.split("?")[1]
    except IndexError:
        return params
    p_str = p_str.split(" ")[0]
    p_str = p_str.split('&')
    for p in p_str:
        k, v = p.split("=")
        params[k] = v
    return params


def median(lst):
    if not lst:
        return 0
    sortedLst = sorted(lst)
    lstLen = len(lst)
    index = (lstLen - 1) // 2

    if (lstLen % 2):
        return sortedLst[index]
    else:
        return (sortedLst[index] + sortedLst[index + 1]) / 2.0


def load_json(f):
    with open(f, 'r') as infile:
        return json.load(infile)



from __future__ import print_function
import csv
import os
import sys
import logging
from mongoengine import connect
from mongoengine.connection import ConnectionError
from pymongo.errors import ServerSelectionTimeoutError
from Bio.Seq import Seq

from atlas.panelgeneration import AlleleGenerator
from atlas.panelgeneration import make_variant_probe
from ga4ghmongo.schema import Variant
from atlas.utils import split_var_name
from atlas.annotation.genes import GeneAminoAcidChangeToDNAVariants
from atlas.panelgeneration.models import Mutation

# logging = logging.getLogger(__name__)
# logging.setLevel(level=logging.DEBUG)


def run(parser, args):
    DB = connect('atlas-%s' % (args.db_name))
    if DB is not None:
        try:
            Variant.objects()
        except (ServerSelectionTimeoutError, ConnectionError):
            DB = None
            logging.warning(
                "Could not connect to database. Continuing without using genetic backgrounds")
    mutations = []
    reference = os.path.basename(args.reference_filepath).split('.fa')[0]
    if args.genbank:
        aa2dna = GeneAminoAcidChangeToDNAVariants(
            args.reference_filepath,
            args.genbank)
        if args.file:
            with open(args.file, 'r') as infile:
                reader = csv.reader(infile, delimiter="\t")
                for row in reader:
                    gene, mutation, alphabet = row
                    if alphabet == "DNA":
                        protein_coding_var = False
                    else:
                        protein_coding_var = True
                    for var_name in aa2dna.get_variant_names(
                            gene, mutation, protein_coding_var):
                        mutations.append(
                            Mutation(reference=reference,
                                     var_name=var_name,
                                     gene=aa2dna.get_gene(gene),
                                     mut=mutation))
        else:
            for variant in args.variant:
                gene, mutation = variant.split("_")
                for var_name in aa2dna.get_variant_names(gene, mutation):
                    mutations.append(
                        Mutation(reference=reference,
                                 var_name=var_name,
                                 gene=gene,
                                 mut=mutation))
    else:
        if args.file:
            with open(args.file, 'r') as infile:
                reader = csv.reader(infile)
                for row in reader:
                    mutations.append(
                        Mutation(
                            reference=reference,
                            var_name=row[0]))
        else:
            mutations.extend(Mutation(reference=reference, var_name=v)
                             for v in args.variants)

    al = AlleleGenerator(
        reference_filepath=args.reference_filepath,
        kmer=args.kmer)
    for mut in mutations:
        variant_panel = make_variant_probe(al, mut.variant, args.kmer, DB=DB)
        if variant_panel is not None:
            if mut.gene:
                sys.stdout.write(
                    ">ref-%s?num_alts=%i&gene=%s&mut=%s&ref=%s\n" %
                    (mut.variant.var_name, len(
                        variant_panel.alts), mut.gene.name, mut.mut, mut.reference))
            else:
                sys.stdout.write(
                    ">ref-%s?num_alts=%i\n" %
                    (mut.variant.var_name, len(
                        variant_panel.alts)))
            sys.stdout.write("%s\n" % variant_panel.ref)
            for a in variant_panel.alts:
                sys.stdout.write(">alt-%s\n" % mut.mut)
                sys.stdout.write("%s\n" % a)
        else:
            logging.warning(
                "All variants failed for %s_%s - %s" %
                (mut.gene, mut.mut, mut.variant))



from __future__ import print_function
import os
import argparse


class ArgumentParserWithDefaults(argparse.ArgumentParser):

    def __init__(self, *args, **kwargs):
        super(ArgumentParserWithDefaults, self).__init__(*args, **kwargs)
        self.add_argument(
            "-q",
            "--quiet",
            help="do not output warnings to stderr",
            action="store_true",
            dest="quiet")

DEFAULT_KMER_SIZE = os.environ.get("KMER_SIZE", 21)
DEFAULT_DB_NAME = os.environ.get("DB_NAME", "atlas")

sequence_parser_mixin = argparse.ArgumentParser(add_help=False)
sequence_parser_mixin.add_argument(
    'sample',
    type=str,
    help='sample id')
sequence_parser_mixin.add_argument(
    'seq',
    type=str,
    help='sequence files (fastq or bam)',
    nargs='+')
sequence_parser_mixin.add_argument(
    '-k',
    '--kmer',
    metavar='kmer',
    type=int,
    help='kmer length (default:21)',
    default=DEFAULT_KMER_SIZE)
sequence_parser_mixin.add_argument(
    '--tmp',
    help='tmp directory (default: /tmp/)',
    default="/tmp/")
sequence_parser_mixin.add_argument(
    '--skeleton_dir',
    help='directory for skeleton binaries',
    default="atlas/data/skeletons/")
sequence_parser_mixin.add_argument(
    '--mccortex31_path',
    help='Path to mccortex31',
    default="mccortex31")



from atlas.typing.models.base import ProbeCoverage
from atlas.typing.models.presence import SequenceProbeCoverage
from atlas.typing.models.panel import Panel
from atlas.typing.models.variant import VariantProbeCoverage
from atlas.typing.typer.presence import PresenceTyper
from atlas.typing.typer.variant import VariantTyper
from atlas.typing.typer.genotyper import Genotyper
from atlas.typing.typer.genotyper import CoverageParser



# Read the kmer counts into a hash
from atlas.utils import check_args
from atlas.typing import Genotyper
from atlas.typing import CoverageParser
from atlas.metagenomics import AMRSpeciesPredictor
from atlas.version import __version__

from pprint import pprint
import json


def run(parser, args):
    args = parser.parse_args()
    # check_args(args)

    # panels = ["tb-species-extended"]
    verbose = True
    cp = CoverageParser(
        sample=args.sample,
        panel_file_paths=args.panels,
        seq=args.seq,
        kmer=args.kmer,
        force=args.force,
        verbose=verbose,
        tmp_dir=args.tmp,
        skeleton_dir=args.skeleton_dir,
        mccortex31_path=args.mccortex31_path)
    cp.run()

    if args.expected_depth is None:
        args.expected_depth = cp.estimate_depth()

    base_json = {args.sample: {}}
    base_json[args.sample]["panels"] = args.panels
    base_json[args.sample]["files"] = args.seq
    base_json[args.sample]["kmer"] = args.kmer
    base_json[args.sample]["version"] = __version__
    gt = Genotyper(sample=args.sample, expected_depths=[args.expected_depth],
                   variant_covgs=cp.variant_covgs,
                   gene_presence_covgs=cp.covgs["presence"],
                   base_json=base_json,
                   contamination_depths=[],
                   force_gt=args.force_gt)
    gt.run()
    cp.remove_temporary_files()
    print(json.dumps(gt.out_json, indent=4))



from __future__ import print_function
import os
import json
import csv
import glob
import logging
import subprocess
from copy import copy

from atlas.typing import SequenceProbeCoverage
from atlas.typing import VariantProbeCoverage
from atlas.typing import ProbeCoverage
from atlas.typing import Panel

from atlas.typing.typer.presence import GeneCollectionTyper
from atlas.typing.typer.variant import VariantTyper

from ga4ghmongo.schema import VariantCallSet
from ga4ghmongo.schema import Variant

from atlas.cortex import McCortexRunner

from atlas.utils import get_params
from atlas.utils import split_var_name
from atlas.utils import median

logger = logging.getLogger(__name__)


def max_pnz_threshold(vp):
    t = max(
        100 - 2 * math.floor(float(max([len(alt) for alt in vp.alts])) / 100), 30)
    return t


class CoverageParser(object):

    def __init__(
            self,
            sample,
            panel_file_paths,
            seq,
            kmer,
            force,
            panels=None,
            verbose=True,
            tmp_dir='/tmp/',
            skeleton_dir='atlas/data/skeletons/',
            mccortex31_path="mccortex31"):
        self.sample = sample
        self.seq = seq
        self.kmer = kmer
        self.force = force
        self.covgs = {"variant": {}, "presence": {}}
        self.variant_covgs = self.covgs["variant"]
        self.gene_presence_covgs = self.covgs["presence"]
        self.mc_cortex_runner = None
        self.verbose = verbose
        self.skeleton_dir = skeleton_dir
        self.tmp_dir = tmp_dir
        self.panel_file_paths = panel_file_paths
        self.panels = []
        self.mccortex31_path = mccortex31_path
        for panel_file_path in self.panel_file_paths:
            panel = Panel(panel_file_path)
            self.panels.append(panel)

    def run(self):
        self._run_cortex()
        self._parse_covgs()

    def _run_cortex(self):
        self.mc_cortex_runner = McCortexRunner(
            sample=self.sample,
            panels=self.panels,
            seq=self.seq,
            kmer=self.kmer,
            force=self.force,
            panel_name=self.panel_name,
            tmp_dir=self.tmp_dir,
            skeleton_dir=self.skeleton_dir,
            mccortex31_path=self.mccortex31_path)
        self.mc_cortex_runner.run()

    def estimate_depth(self):
        depth = []
        for variant_coverages in self.variant_covgs.values():
            for variant_covg in variant_coverages:
                if variant_covg.reference_coverage.median_depth > 0:
                    depth.append(variant_covg.reference_coverage.median_depth)
        return median(depth)

    def remove_temporary_files(self):
        self.mc_cortex_runner.remove_temporary_files()

    @property
    def panel_name(self):
        return "-".join([panel.name for panel in self.panels])

    def _parse_summary_covgs_row(self, row):
        try:
            return row[0], int(row[2]), int(row[3]), 100 * float(row[4])
        except ValueError:
            logger.warning("Failed to parse %s" % str(row))
            return row[0], 0, 0, 0.0

    def _parse_covgs(self):
        with open(self.mc_cortex_runner.covg_tmp_file_path, 'r') as infile:
            self.reader = csv.reader(infile, delimiter="\t")
            for row in self.reader:
                allele, median_depth, min_depth, percent_coverage = self._parse_summary_covgs_row(
                    row)
                allele_name = allele.split('?')[0]
                if self._is_variant_panel(allele_name):
                    self._parse_variant_panel(row)
                else:
                    self._parse_seq_panel(row)

    def _is_variant_panel(self, allele_name):
        try:
            alt_or_ref, _id = allele_name.split('-')
            return bool(alt_or_ref)
        except ValueError:
            return False

    def _parse_seq_panel(self, row):
        allele, median_depth, min_depth, percent_coverage = self._parse_summary_covgs_row(
            row)
        probe_coverage = ProbeCoverage(
            percent_coverage=percent_coverage,
            median_depth=median_depth,
            min_depth=min_depth)

        allele_name = allele.split('?')[0]
        params = get_params(allele)
        panel_type = params.get("panel_type", "presence")
        name = params.get('name')
        version = params.get(
            'version',
            '1')
        if panel_type in ["variant", "presence"]:
            sequence_probe_coverage = SequenceProbeCoverage(
                name=name,
                probe_coverage=probe_coverage,
                version=version,
                length=params.get("length"))
            try:
                self.covgs[panel_type][name][version] = sequence_probe_coverage
            except KeyError:
                self.covgs[panel_type][name] = {}
                self.covgs[panel_type][name][version] = sequence_probe_coverage

        else:
            # Species panels are treated differently
            l = int(params.get("length", -1))
            try:
                self.covgs[panel_type][name]["total_bases"] += l
                if percent_coverage > 75 and median_depth > 0:
                    self.covgs[panel_type][name][
                        "percent_coverage"].append(percent_coverage)
                    self.covgs[panel_type][name]["length"].append(l)
                    self.covgs[panel_type][name]["median"].append(median_depth)
            except KeyError:
                if panel_type not in self.covgs:
                    self.covgs[panel_type] = {}
                self.covgs[panel_type][name] = {}
                self.covgs[panel_type][name]["total_bases"] = l
                if percent_coverage > 75 and median_depth > 0:
                    self.covgs[panel_type][name][
                        "percent_coverage"] = [percent_coverage]
                    self.covgs[panel_type][name]["length"] = [l]
                    self.covgs[panel_type][name]["median"] = [median_depth]
                else:
                    self.covgs[panel_type][name]["percent_coverage"] = []
                    self.covgs[panel_type][name]["length"] = []
                    self.covgs[panel_type][name]["median"] = []

    def _parse_variant_panel(self, row):
        allele, reference_median_depth, min_depth, reference_percent_coverage = self._parse_summary_covgs_row(
            row)
        var_name = allele.split('?')[0].split('-')[1]
        params = get_params(allele)
        num_alts = int(params.get("num_alts", 0))
        reference_coverage = ProbeCoverage(
            percent_coverage=reference_percent_coverage,
            median_depth=reference_median_depth,
            min_depth=min_depth)
        alternate_coverages = []
        for i in range(num_alts):
            row = self.reader.next()
            alt_allele, alternate_median_depth, min_depth, alternate_percent_coverage = self._parse_summary_covgs_row(
                row)
            alternate_coverages.append(
                ProbeCoverage(
                    min_depth=min_depth,
                    percent_coverage=alternate_percent_coverage,
                    median_depth=alternate_median_depth))
        variant_probe_coverage = VariantProbeCoverage(
            reference_coverage=reference_coverage,
            alternate_coverages=alternate_coverages,
            var_name=var_name,
            params=params)
        try:
            self.variant_covgs[allele].append(variant_probe_coverage)
        except KeyError:
            self.variant_covgs[allele] = [variant_probe_coverage]


class Genotyper(object):

    """Takes output of mccortex coverages and types"""

    def __init__(
            self,
            sample,
            expected_depths,
            variant_covgs,
            gene_presence_covgs,
            contamination_depths=[],
            base_json={},
            include_hom_alt_calls=False,
            force_gt=False):
        self.sample = sample
        self.variant_covgs = variant_covgs
        self.gene_presence_covgs = gene_presence_covgs
        self.out_json = base_json
        self.expected_depths = expected_depths
        self.contamination_depths = contamination_depths
        self.variant_calls = {}
        self.sequence_calls = {}
        self.include_hom_alt_calls = include_hom_alt_calls
        self.force_gt = force_gt

    def run(self):
        self._type()

    def _type(self):
        self._type_genes()
        self._type_variants()

    def _type_genes(self):
        gt = GeneCollectionTyper(
            expected_depths=self.expected_depths,
            contamination_depths=self.contamination_depths)
        gene_presence_covgs_out = {}
        for gene_name, gene_collection in self.gene_presence_covgs.items():
            self.gene_presence_covgs[gene_name] = gt.type(gene_collection)
            gene_presence_covgs_out[gene_name] = self.gene_presence_covgs[
                gene_name].to_mongo().to_dict()
        self.out_json[self.sample][
            "sequence_calls"] = gene_presence_covgs_out

    def _type_variants(self):
        self.out_json[self.sample]["variant_calls"] = {}
        out_json = self.out_json[self.sample]["variant_calls"]
        gt = VariantTyper(
            expected_depths=self.expected_depths,
            contamination_depths=self.contamination_depths,
            force_gt=self.force_gt)

        for probe_name, probe_coverages in self.variant_covgs.items():
            variant = self._create_variant(probe_name)
            call = gt.type(probe_coverages, variant=variant)
            if sum(
                    call.genotype) > 0 or not call.genotype or self.include_hom_alt_calls:
                self.variant_calls[probe_name] = call
                if variant is not None:
                    tmp_var = copy(call.variant)
                    call.variant = None
                    out_json["-".join(tmp_var.names)
                             ] = call.to_mongo().to_dict()
                    self.variant_calls[probe_name].variant = tmp_var
                else:
                    out_json[probe_name] = call.to_mongo().to_dict()

    def _create_variant(self, probe_name):
        names = []
        params = get_params(probe_name)
        if params.get("mut"):
            names.append("_".join([params.get("gene"), params.get("mut")]))
        var_name = probe_name.split('?')[0].split('-')[1]
        names.append(var_name)
        try:
            # If it's a variant panel we can create a variant
            ref, start, alt = split_var_name(var_name)
            return Variant.create(
                start=start,
                reference_bases=ref,
                alternate_bases=[alt],
                names=names,
                info=params)
        except AttributeError:
            return None



from __future__ import print_function
import logging
from pprint import pprint
import json
import os
from atlas.utils import check_args
from atlas.typing import CoverageParser
from atlas.typing import Genotyper
from atlas.pheno import TBPredictor
from atlas.pheno import StaphPredictor
from atlas.pheno import GramNegPredictor
from atlas.metagenomics import AMRSpeciesPredictor
from atlas.version import __version__
STAPH_PANELS = ["data/panels/staph-species-160227.fasta.gz",
                "data/panels/staph-amr-bradley_2015.fasta.gz"]

GN_PANELS = [
    "data/panels/gn-amr-genes",
    "data/panels/Escherichia_coli",
    "data/panels/Klebsiella_pneumoniae",
    "data/panels/gn-amr-genes-extended"]


def run(parser, args):
    base_json = {args.sample: {}}
    args = parser.parse_args()
    hierarchy_json_file = None
    if args.panel is not None:
        if args.panel == "bradley-2015":
            TB_PANELS = [
                "data/panels/tb-species-160227.fasta.gz",
                "data/panels/tb-amr-bradley_2015.fasta.gz"]
        elif args.panel == "walker-2015":
            TB_PANELS = [
                "data/panels/tb-species-160227.fasta.gz",
                "data/panels/tb-amr-walker_2015.fasta.gz"]

    if not args.species:
        panels = TB_PANELS + GN_PANELS + STAPH_PANELS
        panel_name = "tb-gn-staph-amr"

    elif args.species == "staph":
        panels = STAPH_PANELS
        panel_name = "staph-amr"
        # hierarchy_json_file = "data/phylo/saureus_hierarchy.json"

    elif args.species == "tb":
        panels = TB_PANELS
        panel_name = "tb-amr"
        hierarchy_json_file = "data/phylo/mtbc_hierarchy.json"
    elif args.species == "gn":
        panels = GN_PANELS
        panel_name = "gn-amr"
    logging.info("Running AMR prediction with panels %s" % ", ".join(panels))
    base_json[args.sample]["panels"] = panels
    base_json[args.sample]["files"] = args.seq
    base_json[args.sample]["kmer"] = args.kmer
    base_json[args.sample]["version"] = __version__
    # Get real paths for panels
    panels = [
        os.path.realpath(
            os.path.join(
                os.path.dirname(__file__),
                "..",
                f)) for f in panels]
    if hierarchy_json_file is not None:
        hierarchy_json_file = os.path.realpath(
            os.path.join(
                os.path.dirname(__file__),
                "..",
                hierarchy_json_file))
    # Run Cortex
    cp = CoverageParser(
        sample=args.sample,
        panel_file_paths=panels,
        seq=args.seq,
        kmer=args.kmer,
        force=args.force,
        verbose=False,
        tmp_dir=args.tmp,
        skeleton_dir=args.skeleton_dir,
        mccortex31_path=args.mccortex31_path)
    cp.run()
    # Detect species
    species_predictor = AMRSpeciesPredictor(
        phylo_group_covgs=cp.covgs.get(
            "complex",
            cp.covgs.get(
                "phylo_group",
                {})),
        sub_complex_covgs=cp.covgs.get(
            "sub-complex",
            {}),
        species_covgs=cp.covgs["species"],
        lineage_covgs=cp.covgs.get(
            "sub-species",
            {}),
        base_json=base_json[args.sample],
        hierarchy_json_file=hierarchy_json_file)
    species_predictor.run()

    # ## AMR prediction

    depths = []
    Predictor = None
    if species_predictor.is_saureus_present():
        depths = [species_predictor.out_json["phylogenetics"]
                  ["phylo_group"]["Staphaureus"]["median_depth"]]
        Predictor = StaphPredictor
    elif species_predictor.is_mtbc_present():
        depths = [species_predictor.out_json["phylogenetics"]["phylo_group"][
            "Mycobacterium_tuberculosis_complex"]["median_depth"]]
        Predictor = TBPredictor
    elif species_predictor.is_gram_neg_present():
        Predictor = GramNegPredictor
        try:
            depths = [species_predictor.out_json["phylogenetics"][
                "species"]["Klebsiella_pneumoniae"]["median_depth"]]
        except KeyError:
            depths = [species_predictor.out_json["phylogenetics"]
                      ["species"]["Escherichia_coli"]["median_depth"]]
    # pprint (species_predictor.out_json["phylogenetics"]["species"])
    # Genotype
    q = args.quiet
    args.quiet = True
    if depths:
        gt = Genotyper(sample=args.sample, expected_depths=depths,
                       variant_covgs=cp.variant_covgs,
                       gene_presence_covgs=cp.covgs["presence"],
                       base_json=base_json,
                       contamination_depths=[],
                       include_hom_alt_calls=True)
        gt.run()
    args.quiet = q
    if Predictor is not None:
        predictor = Predictor(variant_calls=gt.variant_calls,
                              called_genes=gt.gene_presence_covgs,
                              base_json=base_json[args.sample])
        predictor.run()
    cp.remove_temporary_files()
    print(json.dumps(base_json, indent=4))



import logging
import os
import csv
from mongoengine import connect
from mongoengine import NotUniqueError
from mongoengine import OperationError
from mongoengine import DoesNotExist
from pymongo import MongoClient

from atlas.utils import check_args
from ga4ghmongo.schema import ReferenceSet
from ga4ghmongo.schema import Reference
from atlas._vcf import VCF

"""Adds variants to the database"""

logger = logging.getLogger(__name__)
client = MongoClient()


def is_record_valid(record):
    valid = True
    for sample in record.samples:
        if sample["GT"] is None:
            valid = False
        else:
            if sum([int(i) for i in sample['GT'].split('/')]) < 2:
                valid = False
        try:
            if sample["GT_CONF"] <= 1:
                valid = False
        except AttributeError:
            pass
    return valid


def get_genotype_likelihood(sample):
    try:
        genotype_likelihood = sample['GT_CONF']
    except AttributeError:
        genotype_likelihood = sample['GQ']
    return genotype_likelihood


def run(parser, args):
    args = parser.parse_args()
    args = check_args(args)
    if args.quiet:
        logger.setLevel(logging.ERROR)
    else:
        logger.setLevel(logging.INFO)
    DBNAME = 'atlas-%s' % (args.db_name)
    db = client[DBNAME]
    connect(DBNAME)
    logger.debug("Using DB %s" % DBNAME)
    try:
        reference_set = ReferenceSet.objects.get(name=args.reference_set)
    except DoesNotExist:
        reference_set = ReferenceSet.create_and_save(name=args.reference_set)
        # Hack
    try:
        reference = Reference.create_and_save(
            name=args.reference_set,
            reference_sets=[reference_set],
            md5checksum="NA")
    except:
        pass
    vcf = VCF(args.vcf, reference_set.id, method=args.method, force=args.force)
    vcf.add_to_database()



from atlas.typing.typer.base import Typer
from atlas.typing.typer.base import MIN_LLK
from ga4ghmongo.schema import SequenceCall
from atlas.stats import log_lik_depth
from atlas.stats import percent_coverage_from_expected_coverage
from math import log


class PresenceTyper(Typer):

    "Initiated with expected depths and contamination depths"

    def __init__(self, expected_depths, contamination_depths=[]):
        super(
            PresenceTyper,
            self).__init__(
            expected_depths,
            contamination_depths)
        if len(expected_depths) > 1:
            raise NotImplementedError("Mixed samples not supported")

    def type(self, sequence_probe_coverage):
        "Takes a single SequenceCoverage object (or child) and returns genotype"
        call = self._type(sequence_probe_coverage)
        return call

    def _type(self, sequence_probe_coverage):
        hom_alt_likelihoods = []
        het_likelihoods = []
        hom_ref_likelihoods = []
        for expected_depth in self.expected_depths:
            hom_alt_likelihoods.append(
                self._hom_alt_likeihood(
                    median_depth=sequence_probe_coverage.median_depth,
                    expected_depth=expected_depth))
            if not self.has_contamination():
                het_likelihoods.append(
                    self._het_likelihood(
                        median_depth=sequence_probe_coverage.median_depth,
                        expected_depth=expected_depth))
            else:
                het_likelihoods.append(MIN_LLK)

            hom_ref_likelihoods.append(
                self._hom_ref_likelihood(
                    median_depth=sequence_probe_coverage.median_depth,
                    expected_depth=expected_depth))

            for contamination_depth in self.contamination_depths:
                hom_alt_likelihoods.append(
                    self._hom_alt_likeihood(
                        median_depth=sequence_probe_coverage.median_depth,
                        expected_depth=expected_depth +
                        contamination_depth))
                # NOTE : _HOM_ALT_LIKEIHOOD is not a typo
                hom_ref_likelihoods.append(
                    self._hom_alt_likeihood(
                        median_depth=sequence_probe_coverage.median_depth,
                        expected_depth=contamination_depth))
            # Posterior
        hom_ref_likelihood = self._log_post_hom_ref(max(hom_ref_likelihoods))
        hom_alt_likelihood = self._log_post_het_or_alt(
            max(hom_alt_likelihoods),
            expected_depth * 0.75,
            sequence_probe_coverage)
        het_likelihood = self._log_post_het_or_alt(
            max(het_likelihoods),
            expected_depth *
            self.minimum_detectable_frequency,
            sequence_probe_coverage)
        likelihoods = [hom_ref_likelihood, het_likelihood, hom_alt_likelihood]
        gt = self.likelihoods_to_genotype(likelihoods)
        info = {
            "copy_number": float(
                sequence_probe_coverage.median_depth) / expected_depth,
            "coverage": sequence_probe_coverage.coverage_dict,
            "expected_depths": self.expected_depths,
            "contamination_depths": self.contamination_depths
        }
        if sum([int(i) for i in gt.split("/")]) > 0:
            info["version"] = sequence_probe_coverage.version
        if sequence_probe_coverage.length is not None:
            info["length"] = sequence_probe_coverage.length

        return SequenceCall.create(
            sequence=None,
            call_set=None,
            genotype=gt,
            genotype_likelihoods=likelihoods,
            info=info)

    def _hom_alt_likeihood(self, median_depth, expected_depth):
        return log_lik_depth(median_depth, expected_depth * 0.75)

    def _het_likelihood(self, median_depth, expected_depth):
        return log_lik_depth(
            median_depth,
            expected_depth *
            self.minimum_detectable_frequency)

    def _hom_ref_likelihood(self, median_depth, expected_depth):
        return log_lik_depth(median_depth, expected_depth * 0.001)

    @property
    def minimum_detectable_frequency(self):
        if self.error_rate < 0.1:
            return 0.05
        else:
            return 0.25

    def _log_post_hom_ref(self, llk):
        return log(1) + llk

    def _log_post_het_or_alt(self, llk, expected_depth, sequence_coverage):
        expected_percentage_coverage = percent_coverage_from_expected_coverage(
            expected_depth)
        minimum_percentage_coverage_required = expected_percentage_coverage * \
            sequence_coverage.percent_coverage_threshold
        if sequence_coverage.percent_coverage > minimum_percentage_coverage_required:
            return self._log_post_hom_ref(llk)
        else:
            return MIN_LLK


class GeneCollectionTyper(Typer):

    """Types a collection of genes returning only the most likely version
        in the collection"""

    def __init__(self, expected_depths, contamination_depths=[]):
        super(
            GeneCollectionTyper,
            self).__init__(
            expected_depths,
            contamination_depths)
        self.presence_typer = PresenceTyper(
            expected_depths, contamination_depths)

    def type(self, sequence_coverage_collection):
        """Types a collection of genes returning the most likely gene version
            in the collection with it's genotype"""
        best_version = self.get_best_version(
            sequence_coverage_collection.values())
        return self.presence_typer.type(best_version)

    def get_best_version(self, sequence_coverages):
        sequence_coverages.sort(key=lambda x: x.percent_coverage, reverse=True)
        current_best_gene = sequence_coverages[0]
        for gene in sequence_coverages[1:]:
            if gene.percent_coverage < current_best_gene.percent_coverage:
                return current_best_gene
            else:
                if gene.min_depth > current_best_gene.min_depth:
                    current_best_gene = gene
                elif gene.min_depth == current_best_gene.min_depth:
                    if gene.median_depth > current_best_gene.median_depth:
                        current_best_gene = gene
        return current_best_gene



from __future__ import print_function

import os
import subprocess
import urllib2


class McCortexRunner(object):

    def __init__(
            self,
            sample,
            panels,
            seq,
            kmer=31,
            force=False,
            panel_name=None,
            tmp_dir='/tmp/',
            skeleton_dir='data/skeletons/',
            mccortex31_path="mccortex31"):
        self.sample = sample
        self.panels = panels
        self.seq = seq
        self.kmer = kmer
        self.force = force
        self._panel_name = panel_name
        self.tmp_dir = tmp_dir
        if skeleton_dir == 'atlas/data/skeletons/':
            skeleton_dir = os.path.realpath(
                os.path.join(
                    os.path.dirname(
                        os.path.realpath(__file__)),
                    "..",
                    skeleton_dir))
        self.skeleton_dir = skeleton_dir
        self.mccortex31_path = mccortex31_path

    def run(self):
        if self.force or not os.path.exists(self.covg_tmp_file_path):
            self._check_panels()
            self._run_cortex()

    def _check_panels(self):
        # If panel does not exists then build it
        for panel in self.panels:
            if not os.path.exists(panel.filepath):
                raise ValueError(
                    "Could not find a panel at %s." %
                    panel.filepath)

    def _run_cortex(self):
        # If ctx binary does not exist then build it
        self._build_panel_binary_if_required()
        # Now get coverage on panel
        self._run_coverage_if_required()

    def _build_panel_binary_if_required(self):
        if not os.path.exists(self.ctx_skeleton_filepath) or self.force:
            if os.path.exists(self.ctx_skeleton_filepath):
                os.remove(self.ctx_skeleton_filepath)
            # panel
            seq_list = self._create_sequence_list()
            cmd = [self.mccortex31_path,
                   "build",
                   "-m 5GB",
                   "-k",
                   str(self.kmer)] + seq_list + [self.ctx_skeleton_filepath]
            # print (cmd)
            subprocess.check_output(cmd)

    def _create_sequence_list(self):
        seq_list = []
        seq_list.extend(["-s", "%s" % self.panel_name[:100]])
        for panel in self.panels:
            seq_list.extend(["-1", panel.filepath])
        return seq_list

    def _run_coverage_if_required(self):
        if not os.path.exists(
                self.ctx_tmp_filepath) or not os.path.exists(
                self.covg_tmp_file_path) or self.force:
            if os.path.exists(self.ctx_tmp_filepath):
                os.remove(self.ctx_tmp_filepath)
            if os.path.exists(self.covg_tmp_file_path):
                os.remove(self.covg_tmp_file_path)
            subprocess.check_output(self.coverages_cmd)
        else:
            # print "Warning: Using pre-built binaries. Run with --force if
            # panel has been updated."
            pass

    @property
    def coverages_cmd(self):
        cmd = [self.mccortex31_path, "geno",
               "-I", self.ctx_skeleton_filepath,
               "-k", str(self.kmer), "-s", self.sample_name,
               "-o", self.covg_tmp_file_path]
        for seq in self.seq:
            cmd.extend(["-1", seq])
        for panel in self.panels:
            cmd.extend(["-c", panel.filepath])
        cmd.append(self.ctx_tmp_filepath)
        return cmd

    @property
    def sample_name(self):
        return "-".join([self.sample, str(self.kmer)])

    @property
    def panel_name(self):
        if self._panel_name is None:
            self._panel_name = "-".join([p.name.replace('/', '-')
                                         for p in self.panels])
        return self._panel_name

    @property
    def sample_panel_name(self):
        return "_".join([self.sample_name, self.panel_name])

    @property
    def ctx_tmp_filepath(self):
        return "%s/%s.ctx" % (self.tmp_dir, self.sample_panel_name)

    @property
    def covg_tmp_file_path(self):
        return "%s/%s.covgs" % (self.tmp_dir, self.sample_panel_name)

    @property
    def ctx_skeleton_filepath(self):
        return os.path.abspath(
            "%s/%s_%i.ctx" %
            (self.skeleton_dir, self.panel_name.replace(
                "/",
                "-")[
                :100],
                self.kmer))

    def remove_temporary_files(self):
        os.remove(self.ctx_tmp_filepath)
        os.remove(self.covg_tmp_file_path)



from setuptools import setup
setup(
    name='mykatlas',
    version='0.0.1',
    packages=[
        'mykatlas',
        'mykatlas.cmds',
        'mykatlas.typing',
        'mykatlas.typing.typer',
        'mykatlas.typing.models',
        'mykatlas.stats',
        'mykatlas.cortex'],
    license='MIT',
    url='http://github.com/phelimb/atlas',
    description='.',
    author='Phelim Bradley',
    author_email='wave@phel.im',
    install_requires=['future','Biopython','mongoengine','pyvcf','ga4ghmongo'],
    entry_points={
            'console_scripts': [
                'atlas = atlas.atlas_main:main',
            ]})



from __future__ import print_function
import os
import sys
import json
import time
import signal
from Bio.Seq import Seq
from atlas.utils import median
from subprocess import Popen, PIPE
from http.server import BaseHTTPRequestHandler
import socketserver
import logging
from pprint import pprint
import copy
logger = logging.getLogger(__name__)


# Adapted from
# https://github.com/mcveanlab/mccortex/blob/master/scripts/mccortex-server.py
# Start server on port 2306, with link and graph files:
#   python mccortex-server.py 2306 --coverages --edges -p l.ctp.gz a.ctx b.ctx
# Query a kmer:
#   curl localhost:2096/CAGTGGCCA
# Response:
#   { "key": "CAGTGGCCA", "colours": [1], "left": "T", "right": "T", "edges": "88", "links": [] }
#


def check_mccortex_alive(proc):
    if proc.poll() is not None:
        logger.info("McCortex quit [%i]" % proc.returncode, file=sys.stdout)
        sys.exit(1)


def query_mccortex(proc, kmer):
    print(kmer, file=proc.stdin)
    proc.stdin.flush()
    check_mccortex_alive(proc)
    line = proc.stdout.readline()
    # Trim off prompt text
    if line[0:2] == "> ":
        line = line[2:len(line)]
    check_mccortex_alive(proc)
    return line

# when we start mccortex we set it to ignore interrupt signal, we handle it.


def preexec_function():
    # Ignore the SIGINT signal by setting the handler to the standard
    # signal handler SIG_IGN.
    signal.signal(signal.SIGINT, signal.SIG_IGN)


class WebServer(object):

    def __init__(self, port, args):
        self.port = port
        self.args = args
        self.mccortex = None
        self.httpd = None

    def start(self):
        self.mccortex = self._start_mccortex()

    def serve(self):
        mccortex = self.mccortex
        logger.debug("Starting server")

        class McCortexHTTPServer(BaseHTTPRequestHandler):

            def log_message(self, format, *args):
                # Remove logging
                pass

            def do_GET(self):
                self.send_response(200)
                self.send_header("Content-type", "text/html")
                self.end_headers()
                if len(self.path) < 4 or len(self.path) > 300:
                    jsonstr = "{\"error\": \"Webserver: bad query\"}\n"
                else:
                    jsonstr = query_mccortex(mccortex, self.path[1:])
                try:
                    self.wfile.write(jsonstr.encode("UTF-8"))
                except UnicodeDecodeError:
                    self.wfile.write("{}".encode("UTF-8"))
        try:
            self.httpd = socketserver.TCPServer(
                ("", self.port), McCortexHTTPServer)
        except Exception as e:
            logger.error("Cannot start HTTP server: %s" % str(e))
            sys.exit()
        try:
            self.httpd.serve_forever()
        except:
            logger.info("Exited server")

    def stop(self):
        if self.httpd:
            self.httpd.server_close()
        self._stop_mccortex()
        logger.info("Closed succesfully.")

    def _start_mccortex(self):
        script_dir = os.path.dirname(os.path.realpath(__file__))
        # Adding two lists together appends one to the other
        try:
            proc = Popen(["mccortex31",
                          "server",
                          "--single-line",
                          "--coverages"] + self.args,
                         stdin=PIPE,
                         stdout=PIPE,
                         universal_newlines=True,
                         preexec_fn=preexec_function)
        except Exception as e:
            logger.error(
                "Couldn't start McCortex is mccortex31 in path? : %s " %
                str(e))
            sys.exit(1)

        # Give test query to check it works
        check_mccortex_alive(proc)
        resp = query_mccortex(proc, "hi")
        return proc

    def _stop_mccortex(self):
        print("q\n", file=self.mccortex.stdin)
        self.mccortex.stdin.close()
        # sleep until process has closed
        while self.mccortex.poll() is None:
            time.sleep(1)
        logger.info("McCortex exited with: %s " % str(self.mccortex.poll()))


class McCortexQuery(object):

    def __init__(self, proc):
        # self.base_url = "http://localhost:%i/" % port
        self.proc = proc

    def query(self, kmer, known_kmers=[]):
        return McCortexQueryResult(
            kmer,
            json.loads(
                query_mccortex(
                    self.proc,
                    kmer)),
            known_kmers=known_kmers)


class McCortexQueryResult(object):

    def __init__(self, query_kmer, data, known_kmers=[]):
        self.query_kmer = query_kmer
        self.data = data
        self.known_kmers = known_kmers

    def __str__(self):
        return str(self.data)

    def __repr__(self):
        return str(self.data)

    def forward(self, suggested_kmer=None):
        forward_kmers = []
        if self.complement:
            for l in self.left:
                forward_kmers.append(
                    str(Seq(l + self.data["key"][:-1]).reverse_complement()))
        else:
            for r in self.right:
                forward_kmers.append(str(self.data["key"][1:] + r))
        if suggested_kmer is not None and suggested_kmer in forward_kmers:
            return [suggested_kmer]
        if len(forward_kmers) > 1 and self.known_kmers:
            _forward_kmers = [
                f for f in forward_kmers if f in self.known_kmers]
            if len(_forward_kmers) > 0:
                forward_kmers = _forward_kmers
            assert len(forward_kmers) > 0
        return forward_kmers

    def reverse(self, suggested_kmer=None):
        reverse_kmers = []
        if self.complement:
            for l in self.right:
                reverse_kmers.append(
                    str(Seq(self.data["key"][1:] + l).reverse_complement()))
        else:
            for r in self.left:
                reverse_kmers.append(r + str(self.data["key"][:-1]))
        if suggested_kmer is not None and suggested_kmer in reverse_kmers:
            return [suggested_kmer]
        if len(reverse_kmers) > 1 and self.known_kmers:
            _reverse_kmers = [
                f for f in reverse_kmers if f in self.known_kmers]
            if len(_reverse_kmers) > 0:
                reverse_kmers = _reverse_kmers
            assert len(reverse_kmers) > 0
        return reverse_kmers

    @property
    def right(self):
        return self.data["right"]

    @property
    def left(self):
        return self.data["left"]

    @property
    def kmer(self):
        return self.data["key"]

    @property
    def depth(self):
        try:
            return int(self.data["colours"][0])
        except ValueError:
            logger.error(
                "parsing edges %s:%s " %
                (self.data["key"], self.data["edges"]))
            return None

    @property
    def complement(self):
        return self.kmer != self.query_kmer


class GraphWalker(object):

    def __init__(self, proc, kmer_size=31, print_depths=False):
        self.mcq = McCortexQuery(proc)
        self.queries = {}
        self.kmer_size = kmer_size
        self.print_depths = print_depths

    def _count_k(self, k, count):
        try:
            count[k] += 1
        except KeyError:
            count[k] = 1
        return count

    def _path_ended(self, paths, i):
        return paths[i]["dna"][-1] == "*"

    def _get_next_kmer_right(self, paths, i):
        return paths[i]["dna"][-self.kmer_size:]

    def _get_next_kmer_left(self, paths, i):
        return paths[i]["dna"][:self.kmer_size]

    def _reached_end_of_path(self, k, end_kmers):
        return k in end_kmers

    def _mark_path_and_ended(self, paths, i):
        paths[i]["dna"] = paths[i]["dna"] + "*"
        return paths

    def _make_query(self, k, known_kmers=[]):
        try:
            q = self.queries[k]
        except KeyError:
            try:
                q = self.mcq.query(k, known_kmers=known_kmers)
                self.queries[k] = q
            except ValueError as e:
                q = None
                logging.error(str(e))
                self.queries[k] = None
        return q

    def _query_is_valid(self, q):
        return q is not None and q.data.get("key")

    def _get_next_kmers_right(
            self,
            i,
            q,
            k,
            paths,
            repeat_kmers,
            known_kmers,
            count):
        kmers = q.forward(suggested_kmer=repeat_kmers.get(k, {}).get(count[k]))
        if q.depth is not None:
            if self.print_depths:
                paths[i]["depth"].append(q.depth)
            if len(kmers) > 1:
                depth = max(q.depth * 0.1, 10)
                kmers = [
                    k for k in kmers if self.mcq.query(
                        k,
                        known_kmers=known_kmers).depth > depth]
        if len(kmers) > 1:
            kmers = [k for k in kmers if k not in paths[i]["dna"]]
        return kmers

    def _get_next_kmers_left(
            self,
            i,
            q,
            k,
            paths,
            repeat_kmers,
            known_kmers=[],
            count={}):
        kmers = q.reverse()
        if q.depth is not None:
            if len(kmers) > 1:
                depth = max(q.depth * 0.1, 10)
                kmers = [
                    k for k in kmers if self.mcq.query(
                        k,
                        known_kmers=known_kmers).depth > depth]
        if len(kmers) > 1:
            kmers = [k for k in kmers if k not in paths[i]["dna"]]
        return kmers

    def _check_if_next_kmers_are_valid(self, kmers, paths, i):
        if len(kmers) < 1:
            paths = self._mark_path_and_ended(paths, i)
        return paths

    def _split_paths(self, i, kmers, paths):
        for j, kmer in enumerate(kmers):
            if j == 0:
                paths[i]["dna"] = paths[i]["dna"] + kmer[-1]
            else:
                paths[sorted(
                    paths.keys())[-j]]["dna"] = paths[sorted(paths.keys())[-j]]["dna"] + kmer[-1]
        return paths

    def _init_paths(self, seed, N_left):
        if self.print_depths:
            return {0: {"N_left": N_left,
                        "dna": seed[:self.kmer_size],
                        "start_kmer": seed[:self.kmer_size],
                        "depth": []}}
        else:
            return {0: {"N_left": N_left,
                        "dna": seed[:self.kmer_size],
                        "start_kmer": seed[:self.kmer_size]}}

    def breath_first_search(self, N, seed, end_kmers=[],
                            known_kmers=[], repeat_kmers={},
                            N_left=0):
        paths = self._init_paths(seed, N_left)
        count = {}
        for _ in range(N_left):
            k = self._get_next_kmer_left(paths, 0)
            q = self._make_query(k)
            if self._query_is_valid(q):
                kmers = self._get_next_kmers_left(
                    0,
                    q,
                    k,
                    paths,
                    repeat_kmers,
                    known_kmers,
                    count)
                if not len(kmers) == 1:
                    break
                else:
                    # Add left
                    paths[0]["dna"] = kmers[0][0] + paths[0]["dna"]
        #
        #
        for _ in range(N - N_left):
            for i in paths.keys():
                if not self._path_ended(paths, i):
                    k = self._get_next_kmer_right(paths, i)
                    count = self._count_k(k, count)
                    if self._reached_end_of_path(k, end_kmers):
                        paths = self._mark_path_and_ended(paths, i)
                    else:
                        q = self._make_query(k, known_kmers)
                        if self._query_is_valid(q):
                            kmers = self._get_next_kmers_right(
                                i,
                                q,
                                k,
                                paths,
                                repeat_kmers,
                                known_kmers,
                                count)
                            paths = self._check_if_next_kmers_are_valid(
                                kmers,
                                paths,
                                i)
                            # Create new paths
                            paths = self.create_new_paths(
                                paths,
                                i,
                                kmers,
                                origin=q.data.get("key"))
                            paths = self._split_paths(i, kmers, paths)
                        else:
                            paths = self._mark_path_and_ended(paths, i)

            # DEBUG prints progress every N bps
            # if i % 250 == 0:
            #     with open("/tmp/paths.json", "w") as outfile:
            #         json.dump(paths, outfile)
            ###

        keep_paths = {}
        for k, v in paths.items():
            v["len_dna"] = len(v["dna"])
            v["prot"] = str(Seq(v["dna"].rstrip("*")).translate(11))
            v["len_prot"] = len(v["prot"])
            if self.print_depths:
                paths[i]["median_depth"] = median(paths[i]["depth"])
                paths[i]["min_non_zero_depth"] = min(paths[i]["depth"])
                paths[i]["depth"] = "-".join([str(x)
                                              for x in paths[i]["depth"]])

            if v["len_dna"] == N + 1 and v["dna"][-1] == "*":
                keep_paths[k] = v
        return keep_paths.values()

    def create_new_paths(self, paths, i, kmers, origin):
        num = len(kmers) - 1
        num_paths = max(paths.keys())
        if num > 0:
            logger.debug("Branch point")
            logger.debug("Origin %s" % origin)
            logger.debug("Options %s" % ",".join(kmers))
        for j in range(num):
            paths[num_paths + j + 1] = copy.copy(paths[i])
            paths[num_paths + j + 1]["start_kmer"] = kmers[1]
            if kmers[1] in [d["start_kmer"] for d in paths.values()]:
                print (kmers, paths)
                raise ValueError("Going around in circles?")
        return paths


# def main():
#     if len(sys.argv) < 3 or not sys.argv[1].isdigit():
#         print("usage: %s <port> [mccortex args]" % (sys.argv[0]))
#         sys.exit(-1)

#     port = int(sys.argv[1])
#     args = sys.argv[2:]

#     start_web_server(port, args)

# if __name__ == '__main__':
#     main()





class SequenceProbeCoverage(object):

    def __init__(
            self,
            name,
            probe_coverage,
            percent_coverage_threshold=30,
            version=1,
            length=None):
        self.name = name
        self.probe_coverage = probe_coverage
        self.percent_coverage_threshold = percent_coverage_threshold
        self.version = version
        self.length = length

    @property
    def median_depth(self):
        return self.probe_coverage.median_depth

    @property
    def percent_coverage(self):
        return self.probe_coverage.percent_coverage

    @property
    def min_depth(self):
        return self.probe_coverage.min_depth

    @property
    def coverage_dict(self):
        return self.probe_coverage.coverage_dict
